open Prefs
open Debug
open Print
open Heap
open Terms
open Rewriting
open Checker
open Learning

type orientation_t = 
    LeftToRight | RightToLeft | Either | Neither
;;

let axioms = Hashtbl.create 1;;

let true_const, false_const = T("_true",[]), T("_false",[]);;

let axiom_schemas = ref [(V("X",0), true_const); (V("X",0), false_const)];;

let add_axiom (l',r') =
  let l,r = canon_rule (l',r') in
    if Hashtbl.mem axioms (r,l) then
      failwith "Axiom already oriented"
    else Hashtbl.add axioms (l,r) ()
;;

let is_axiom (l,r) = 
  (r = true_const && l <> false_const) || 
    (r = false_const && l <> true_const) ||
    if !trust_axiom_orientation then 
      Hashtbl.mem axioms (canon_rule (l,r))
    else false
;;

let axiom_filter (l,r) = function
    Neither  -> Neither
  | LeftToRight -> if is_axiom (r,l) then Neither else LeftToRight
  | RightToLeft -> if is_axiom (l,r) then Neither else RightToLeft
  | Either -> 
      if is_axiom (l,r) then LeftToRight
      else if is_axiom (r,l) then RightToLeft
      else Either
;;

let orient_all (s,t) rules = 
  orientations := !orientations + 2;
  let status = (* sufficient conditions for non-termination *)
    (* s m.w. t then s -> t is non-term. *)
    if try_match_subterm (s,t) then 
      (register_nonterminating_trs [s,t]; (* is this necessary? *)
       if try_match_subterm(t,s) then 
         (register_nonterminating_trs [t,s];
          Neither) 
       else 
         RightToLeft)
    else if try_match_subterm (t,s) then 
      (register_nonterminating_trs [t,s];
       if try_match_subterm(s,t) then 
         (register_nonterminating_trs [s,t];
          Neither) 
       else LeftToRight)
        (* rules with new variables in t are non-terminating *)
    else
      let s_subset_t = subset_vars s t in
      let t_subset_s = subset_vars t s in
        if s_subset_t && not t_subset_s then
          (register_nonterminating_trs [s,t];
           RightToLeft)
        else if not s_subset_t && t_subset_s then
          (register_nonterminating_trs [t,s];
           LeftToRight)
        else if not s_subset_t && not t_subset_s then
          (register_nonterminating_trs [s,t];
           register_nonterminating_trs [t,s];
           Neither)
        else (* ask aprove *)
          Either
  in
  let status' = 
    axiom_filter (s,t) status
  in
(*   let status'',lr_t,rl_t =  *)
    let all_terms = s::t::
      (List.fold_left (fun xs (s,t) -> s::t::xs) [] rules) in
    let vars = find_vars_all all_terms in
      match status' with
        | Neither -> 
            Neither,max_float,max_float
        | LeftToRight -> 
            let result,time = 
              if recall_nonterminating_trs ((s,t)::rules) then 
                Neither,max_float
              else if recall_terminating_trs ((s,t)::rules) then
                LeftToRight,0.
              else
                let term,time = is_terminating ((s,t)::rules) vars in
                  if term then 
                    (register_terminating_trs ((s,t)::rules); 
                     LeftToRight,time)
                  else Neither,max_float
            in
              result,time,max_float
        | RightToLeft -> 
            let result,time = 
              if recall_nonterminating_trs ((t,s)::rules) then 
                Neither,max_float
              else if recall_terminating_trs ((t,s)::rules) then
                LeftToRight,0.
              else
                let term,time = is_terminating ((t,s)::rules) vars in
                  if term then 
                    (register_terminating_trs ((t,s)::rules); 
                     RightToLeft,time)
                  else Neither,max_float
            in
              result,max_float,time
        | Either ->
            let left_recall_nt, right_recall_nt =
              recall_nonterminating_trs ((s,t)::rules),
              recall_nonterminating_trs ((t,s)::rules)
            in
              if left_recall_nt && right_recall_nt then
                Neither,max_float,max_float
              else if left_recall_nt && not right_recall_nt then 
                let checker_rvs = [(((t,s)::rules),vars)] in
                  match are_terminating checker_rvs with
                      [true,right_time] -> 
                        RightToLeft,max_float,right_time
                    | [false,right_time] ->
                        Neither,max_float,right_time
                    | _ -> failwith "Extra checker results"
              else if not left_recall_nt && right_recall_nt then
                let checker_rvs = [(((t,s)::rules),vars)] in
                  match are_terminating checker_rvs with
                      [true,right_time] -> 
                        RightToLeft,max_float,right_time
                    | [false,right_time] ->
                        Neither,max_float,right_time
                    | _ -> failwith "Extra checker results"
              else (* not left_recall_nt && not right_recall_nt *)
                let left_recall_t, right_recall_t = 
                  recall_terminating_trs ((s,t)::rules),
                  recall_terminating_trs ((t,s)::rules)
                in
                  if left_recall_t && right_recall_t then
                    Either,0.,0.
                  else if left_recall_t && not right_recall_t then 
                    match are_terminating [((t,s)::rules),vars] with
                        [true,time] -> Either,0.,time
                      | [false,time] -> LeftToRight,0.,time
                      | _ -> failwith "Extra checker results (1)"
                  else if not left_recall_t && right_recall_t then
                    match are_terminating [((s,t)::rules),vars] with
                        [true,time] -> Either,time,0.
                      | [false,time] -> RightToLeft,time,0.
                      | _ -> failwith "Extra checker results (2)"
                  else (* not left_recall_t && not right_recall_t *)
                    let checker_rvs = [(((s,t)::rules),vars);(((t,s)::rules),vars)] in
                      match are_terminating checker_rvs with
                          [(true,left_time);(true,right_time)] -> 
                            Either,left_time,right_time
                        | [(true,left_time);(false,right_time)] -> 
                            LeftToRight,left_time,right_time
                        | [(false,left_time);(true,right_time)] -> 
                            RightToLeft,left_time,right_time
                        | [(false,left_time);(false,right_time)] -> 
                            Neither,left_time,right_time
                        | _ -> failwith "Incomplete checker results"
              
                          

(*           let lr_start_t = Unix.gettimeofday() in *)
(*           let left =  *)
(*             if recall_nonterminating_trs ((s,t)::rules) then false *)
(*             else *)
(*               recall_terminating_trs ((s,t)::rules) ||  *)
(*                 is_terminating ((s,t)::rules) vars  *)
(*           in *)
(*           let lr_t = (Unix.gettimeofday()) -. lr_start_t in *)
(*           let rl_start_t = Unix.gettimeofday() in *)
(*             begin *)
(*               let right =  *)
(*                 if recall_nonterminating_trs ((t,s)::rules) then false *)
(*                 else  *)
(*                   recall_terminating_trs ((t,s)::rules) ||  *)
(*                     is_terminating ((t,s)::rules) vars  *)
(*               in *)
(*               let rl_t = (Unix.gettimeofday()) -. rl_start_t in *)
(*               let final_status =  *)
(*                 match left,right with *)
(*                     true,true -> Either *)
(*                   | false,false -> Neither *)
(*                   | true,false -> LeftToRight *)
(*                   | false,true -> RightToLeft *)
(*               in final_status,lr_t,rl_t *)
(*             end *)
(*   in *)
(*     status'',lr_t,rl_t *)
;;








