open Prefs
open Debug
open Terms
open Print

(* rewrite: (term * term) list -> term -> term *)
let rec rewrite _l t = 
  match _l with 
      [] -> None
    | ((l,r)::_R) -> 
        match match_with(l,t) with
            None -> rewrite _R t
          | Some s -> 
              let s' = lift s r in
                Some s'
;;

let rec norm _R t =
  match t with 
      V x -> V x
    | (T(f,ts)) ->
        let u = T(f, List.map (norm _R) ts) in 
          match rewrite _R u with 
            | None -> u
            | Some u' -> norm _R u'
;;

let normalize _R t = 
  debug_if_l TRACE (!debug_rewriting) (lazy "Normalizing with rules: "); 
  debug_if_l TRACE (!debug_rewriting) (lazy (string_of_trs _R)); 
  let t' = norm _R t in
    debug_if_l TRACE (!debug_rewriting) 
      (lazy ((string_of_term t) ^ " -->! " ^ (string_of_term t')));
    t'
;;

let rec narrow _R (s,t) = 
  let rec _narrow _l t = 
    match _l with 
      | [] -> None
      | (l,r)::_R ->
          match unify(l,t) with 
              None -> _narrow _R t
            | Some sigma ->
                let t' = lift sigma r in
                  Some (t',sigma)
  in
  let rec _narrow_args _R t0s t1s = 
    match t1s with 
      [] -> None
    | t::ts -> 
        match _narrow_term _R t with 
            None -> _narrow_args _R (t0s@[t]) ts
          | Some (t',sigma) -> Some (t0s@[t']@ts, sigma)
  and _narrow_term _R _t = 
      match _t with 
      V x -> None
    | T(f,ts) as u ->
        match _narrow _R u with 
            None -> 
              begin 
                match _narrow_args _R [] ts with
                    None -> None
                  | Some (ts',sigma) -> 
                      Some ((T(f,ts')),sigma)
              end
          | Some (t',sigma) -> Some (t',sigma)
  in

  let rec narrow_star _R (s,t) = 
    match _narrow_term _R s with 
        None -> 
          begin
            match _narrow_term _R t with
                None -> (s,t)
              | Some (t',sigma) -> narrow_star _R 
                  (lift sigma s,lift sigma t')
          end
      | Some (s',sigma) -> narrow_star _R (lift sigma s',lift sigma t)
  in

    debug_if_l TRACE (!debug_narrowing) (lazy "Narrowing with rules: "); 
    debug_if_l TRACE (!debug_narrowing) (lazy (string_of_trs _R)); 

    let m = 1 + maxs 
      (List.map (fun (l,r) -> max (maxindex l) (maxindex r)) _R)
    in
    let s',t' = narrow_star _R (rename m s,rename m t) in
      debug_if_l TRACE (!debug_narrowing) 
        (lazy ((string_of_term s) ^ " -->N " ^ (string_of_term s')));
      debug_if_l TRACE (!debug_narrowing) 
        (lazy ((string_of_term t) ^ " -->N " ^ (string_of_term t')));
      (s',t')
      
;;
            




